/*

Copyright 2019, dettus@dettus.net

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


*/

//#define	DEBUG_PRINT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "linea.h"
#include "gfxloader.h"
#include "vm68k_datatypes.h"
#include "vm68k_macros.h"

#define	MAGICVALUE	0x42696e61	// ="Lina"


typedef struct _tProperties
{
	tVM68k_ubyte unknown1[5];
	tVM68k_ubyte flags1;
	tVM68k_ubyte flags2;
	tVM68k_ubyte unknown2;
	tVM68k_uword parentobject;	
	tVM68k_ubyte unknown3[2];
	tVM68k_uword endflags;
} tProperties;

typedef struct _tLineA
{
	tVM68k_ulong magic;
	tVM68k_ubyte	version;

	// pointers to important memory sections.
	tVM68k_ubyte* 	pMem;
	tVM68k_ulong	memsize;
	tVM68k_ulong	codesize;
	tVM68k_ubyte*	pStrings1;
	tVM68k_ulong	string1size;
	tVM68k_ubyte*	pStrings2;
	tVM68k_ulong	string2size;
	tVM68k_ubyte*	pDict;
	tVM68k_ulong	dictsize;	
	tVM68k_ulong	decsize;	
	tVM68k_ubyte*	pUndo;
	tVM68k_ulong	undosize;
	tVM68k_slong	undopc;

	// pointers to the callback functions
	cbLineAOutputChar	pcbOutputChar;
	void* 			contextOutputChar;
	cbLineAOutputString	pcbOutputString;
	void*			contextOutputString;
	cbLineAInputString	pcbInputString;
	void*			contextInputString;
	cbLineADrawPicture	pcbDrawPicture;
	void*			contextDrawPicture;
	cbLineASaveGame		pcbSaveGame;
	void*			contextSaveGame;
	cbLineALoadGame		pcbLoadGame;
	void*			contextLoadGame;

	// persistent memory for some A0xx instructions.
	tVM68k_slong	random_state;
	tVM68k_uword	properties_offset;
	tVM68k_uword	linef_subroutine;			// version >0
	tVM68k_slong	byteidx;
	tVM68k_ubyte	bitidx;

	// input buffer queue.
	char inputbuf[256];
	int level;
	int used;

#define	MAXMAGBUF	183916
#define	MAXGFXBUF	2534113
	// buffer for everything. Personally, I do not like restricting it. 
	// on the other hand: I do not like mallocs spread around the code either.
	tVM68k_ubyte magbuf[MAXMAGBUF];
	tVM68k_ulong magsize;
	tVM68k_ubyte gfxbuf[MAXGFXBUF];
	tVM68k_ulong gfxsize;

	tPicture picture;
} tLineA;

tVM68k_ulong lineA_getrandom(tLineA* pLineA)
{
	static int wtf=0;
	//if (wtf==0) pLineA->random_state=1554745337;	// the pawn
	if (wtf==0) pLineA->random_state=1555620022;	// the guild of thieves
	wtf=1;
	pLineA->random_state*=1103515245ull;
	pLineA->random_state+=12345ull;
	return pLineA->random_state&0x7fffffff;
}
int lineA_savegame(tLineA* pLineA,tVM68k* pVM68k,char* filename)
{
	FILE *f;
	unsigned char header[16]="dMagnetic V0000";
	unsigned char savelong[4];
	unsigned char saveword[2];
	int n;
	int i;

	f=fopen(filename,"wb");
// 16 bytes header
	n=fwrite(header,sizeof(char),16,f);
// write the full shared memory.
	WRITE_INT32BE(savelong,0,pVM68k->memsize);// start with the size in bytes, BIG endian
	n+=fwrite(savelong,sizeof(char),4,f);
	n+=fwrite(pVM68k->pMem,sizeof(char),pVM68k->memsize,f);
// write the status registers of the 68K core
	WRITE_INT32BE(savelong,0,pVM68k->pcr);n+=fwrite(savelong,sizeof(char),4,f);
	WRITE_INT16BE(saveword,0,pVM68k->sr);n+=fwrite(saveword,sizeof(char),2,f);
	for (i=0;i<8;i++)
	{
		WRITE_INT32BE(savelong,0,pVM68k->a[i]);n+=fwrite(savelong,sizeof(char),4,f);
	}
	for (i=0;i<8;i++)
	{
		WRITE_INT32BE(savelong,0,pVM68k->d[i]);n+=fwrite(savelong,sizeof(char),4,f);
	}
// write the persistant memory from the LINEA core
	WRITE_INT32BE(savelong,0,pLineA->random_state);n+=fwrite(savelong,sizeof(char),4,f);
	WRITE_INT16BE(saveword,0,pLineA->properties_offset);n+=fwrite(saveword,sizeof(char),2,f);
	WRITE_INT32BE(savelong,0,pLineA->byteidx);n+=fwrite(savelong,sizeof(char),4,f);
	n+=fwrite(&pLineA->bitidx,sizeof(char),1,f);	
	fclose(f);
	return n;	
}
int lineA_loadgame(tLineA* pLineA,tVM68k* pVM68k,char* filename)
{
	FILE *f;
	unsigned char header[16]="dMagnetic V0000";
	unsigned char loadlong[4];
	unsigned char loadword[2];
	int n;
	int i;

	f=fopen(filename,"rb");
// 16 bytes header
	n=fread(header,sizeof(char),16,f);
// write the full shared memory.
	n+=fread(loadlong,sizeof(char),4,f);pVM68k->memsize=READ_INT32BE(loadlong,0);
	n+=fread(pVM68k->pMem,sizeof(char),pVM68k->memsize,f);	// FIXME: yeah, i know. 
// write the status registers of the 68K core
	n+=fread(loadlong,sizeof(char),4,f);pVM68k->pcr=READ_INT32BE(loadlong,0);
	n+=fread(loadword,sizeof(char),2,f);pVM68k->sr=READ_INT16BE(loadword,0);
	for (i=0;i<8;i++)
	{
		n+=fread(loadlong,sizeof(char),4,f);pVM68k->a[i]=READ_INT32BE(loadlong,0);
	}
	for (i=0;i<8;i++)
	{
		n+=fread(loadlong,sizeof(char),4,f);pVM68k->d[i]=READ_INT32BE(loadlong,0);
	}
// write the persistant memory from the LINEA core
	n+=fread(loadlong,sizeof(char),4,f);pLineA->random_state=READ_INT32BE(loadlong,0);
	n+=fread(loadword,sizeof(char),2,f);pLineA->properties_offset=READ_INT16BE(loadword,0);
	n+=fread(loadlong,sizeof(char),4,f);pLineA->byteidx=READ_INT32BE(loadlong,0);
	n+=fread(&pLineA->bitidx,sizeof(char),1,f);	
	fclose(f);
	return n;	
}
int lineA_parsegamefiles(tLineA* pLineA,void* pMag,int magsize,void* pGfx,int gfxsize)
{
	int idx;
	// first: store the mag and gfx data
	if (pMag==NULL) return LINEA_NOK_INVALID_PTR;
	if (magsize>MAXMAGBUF) return LINEA_NOK_NOT_ENOUGH_MEMORY;
	memcpy(pLineA->magbuf,pMag,magsize);pLineA->magsize=magsize;

	// lets start with the header.
	// @0   4 bytes "MaSc"
	// @4   12 bytes TODO
	// @13  1 byte version
	// @14  4 bytes codesize
	// @18  4 bytes string1size
	// @22  4 bytes string2size
	// @26  4 bytes dictsize
	// @30  4 bytes decsize
	// @34  4 bytes undosize
	// @38  4 bytes undopc

	// ----------
	// @42   codesize bytes   code
	// @42+codesize  string1size bytes...
	// @...  string2size
	// @...  dictsize
	// @...  undo

	if (pLineA->magbuf[0]!='M' || pLineA->magbuf[1]!='a' || pLineA->magbuf[2]!='S' || pLineA->magbuf[3]!='c') return LINEA_NOK_INVALID_PARAM;
//	since the magic word seemed to be okay, start reading the header.
	pLineA->version=pLineA->magbuf[13];
	pLineA->codesize=READ_INT32BE(pLineA->magbuf,14);
	pLineA->string1size=READ_INT32BE(pLineA->magbuf,18);
	pLineA->string2size=READ_INT32BE(pLineA->magbuf,22);
	pLineA->dictsize=READ_INT32BE(pLineA->magbuf,26);
	pLineA->decsize=READ_INT32BE(pLineA->magbuf,30);
	pLineA->undosize=READ_INT32BE(pLineA->magbuf,34);
	pLineA->undopc=READ_INT32BE(pLineA->magbuf,38);

	/////////////////////////////////////
	// the code section has to be copied into the shared memory section.
	// because the same section is writable. first: drain the memory.
	memset(pLineA->pMem,0,pLineA->memsize);
	if (pLineA->memsize<pLineA->codesize) return LINEA_NOK_NOT_ENOUGH_MEMORY;


	idx=42;	// the header held 42 bytes.
	memcpy(pLineA->pMem,&pLineA->magbuf[idx],pLineA->codesize);	idx+=pLineA->codesize;


	// every other sections is just read-only. they can stay where they are.
	// for conveniance reasons, i am adding the pointers.
	pLineA->pStrings1=&pLineA->magbuf[idx];idx+=pLineA->string1size;
	pLineA->pStrings2=&pLineA->magbuf[idx];idx+=pLineA->string2size;
	pLineA->pDict=&pLineA->magbuf[idx];idx+=pLineA->dictsize;
	pLineA->pUndo=&pLineA->magbuf[idx];idx+=pLineA->undosize;
	
	pLineA->gfxsize=gfxsize;
	if (pGfx!=NULL)
	{
		memcpy(pLineA->gfxbuf,pGfx,gfxsize);
	} 
		
	return LINEA_OK;
}
// the purpose of this function is to load the properties for a specific object.
int lineA_loadproperties(tLineA* pLineA,tVM68k* pVM68k,tVM68k_uword objectnum,tVM68k_ulong* retaddr,tProperties* pProperties)
{
	tVM68k_ulong addr;
	int i;
	addr=pLineA->properties_offset+14*objectnum;
	for (i=0;i<5;i++)
	{
		pProperties->unknown1[i]=pVM68k->pMem[addr+i];
	}	
	pProperties->flags1=pVM68k->pMem[addr+5];
	pProperties->flags2=pVM68k->pMem[addr+6];
	pProperties->unknown2=pVM68k->pMem[addr+7];
	pProperties->parentobject=READ_INT16BE(pVM68k->pMem,addr+8);
	for (i=0;i<2;i++)
	{
		pProperties->unknown3[i]=pVM68k->pMem[addr+i+10];
	}
	pProperties->endflags=READ_INT16BE(pVM68k->pMem,addr+12);
	if (retaddr!=NULL) *retaddr=addr;
	return LINEA_OK;	
}

int lineA_getsize(int* size)
{
	if (size==NULL) return LINEA_NOK_INVALID_PTR;
	*size=sizeof(tLineA);
	return LINEA_OK;
}
int lineA_substitute_aliases(void* hLineA,unsigned short* opcode)
{
	tVM68k_uword inst;
	inst=*opcode;
	if ((inst&0xfe00)==0xA400) {inst&=0x01ff;inst|=0x6100;}	// BSR
	if ((inst&0xfe00)==0xA200) {inst=0x4e75;}	// RTS
	if ((inst&0xfe00)==0xA600) {inst&=0x01ff;inst|=0x4a00;}	// TST
	if ((inst&0xfe00)==0xA800) {inst&=0x01ff;inst|=0x4800;}	// MOVEM, register to memory (=0x4800)
	if ((inst&0xfe00)==0xAA00) {inst&=0x01ff;inst|=0x4C00;}	// MOVEM, memory to register (=0x4C00)

	*opcode=inst;
	return LINEA_OK;

}
int lineA_init(void* hLineA,void* pSharedMem,int sharedmemsize,void* pMag,int magsize,void* pGfx,int gfxsize)
{
	tLineA* pLineA=(tLineA*)hLineA;
	int retval;
	if (hLineA==NULL) return LINEA_NOK_INVALID_PTR;
	memset(pLineA,0,sizeof(tLineA));
	pLineA->magic=MAGICVALUE;
	pLineA->random_state=12345;

	pLineA->pMem=pSharedMem;			pLineA->memsize=sharedmemsize;

	pLineA->pcbOutputChar=NULL;	pLineA->contextOutputChar=NULL;
	pLineA->pcbOutputString=NULL;	pLineA->contextOutputString=NULL;
	pLineA->pcbInputString=NULL;	pLineA->contextInputString=NULL;
	pLineA->pcbDrawPicture=NULL;	pLineA->contextDrawPicture=NULL;
	pLineA->pcbSaveGame=NULL;	pLineA->contextSaveGame=NULL;
	pLineA->pcbLoadGame=NULL;	pLineA->contextLoadGame=NULL;
	
	retval=lineA_parsegamefiles(pLineA,pMag,magsize,pGfx,gfxsize);
		
	
	return retval;
}
int lineA_singlestep(void* hLineA,void* hVM68k,unsigned short opcode)
{
	tLineA* pLineA=(tLineA*)hLineA;
	tVM68k* pVM68k=(tVM68k*)hVM68k;

	int retval=LINEA_NOK_UNKNOWN_INSTRUCTION;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pVM68k==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	if ((opcode&0xf000)==0xf000)	// version 1 introduced programmable subroutines.
	{
//		PUSHLONGTOSTACK(pVM68k,&next,next.pcr);
		pVM68k->a[7]-=4;
		WRITE_INT32BE(pVM68k->pMem,pVM68k->a[7],pVM68k->pcr);
		pVM68k->pcr=(pLineA->linef_subroutine)&0xffff;
		return LINEA_OK;
	}
	if ((opcode&0xff00)!=0xa000) return LINEA_NOK_INVALID_PARAM;
	retval=LINEA_OK;
	if ((opcode&0xff)<0xdd)
	{
		lineA_getrandom(pLineA);	// advance the random generator
	}
	switch (opcode)
	{
		case 0xa000:	// getchar
			{
				// the way i implemented the input is this:
				// i have a buffer, inputbuf[256]. holding the last input string.
				// when this instruction is called, one character is read and written into D1.
				// 
				// when the buffer is empty, the callback function for inputs is called first.
				if (pLineA->level==pLineA->used) 
				{
					pLineA->level=0;
					pLineA->used=0;
					if (pLineA->pcbInputString!=NULL)
					{
						pLineA->pcbInputString(pLineA->contextInputString,&pLineA->level,pLineA->inputbuf);	// callback to fill the input buffer.
					} else {
						fprintf(stderr,"INTERNAL ERROR: no input set!\n");
						exit(0);
					}
				}
				if (pLineA->level>pLineA->used)	// still characters in the buffer?
				{
					pVM68k->d[1]=pLineA->inputbuf[pLineA->used];	// yes. take one out
					pLineA->used++;					// increase the read pointer for the next time.
				}
			}
			break;
		case 0xa0ed:	// quit
			{
				retval=LINEA_OK_QUIT;
			}
			break;
		case 0xa0f0:
			{
				//printf("\x1b[1;37;44mLINEA: show picture %d mode %d\x1b[0m\n",pVM68k->d[0],pVM68k->d[1]);
				if (pVM68k->d[1] && pLineA->gfxsize) 
				{
					gfxloader_unpackpic(pLineA->gfxbuf,pLineA->gfxsize,pLineA->version,pVM68k->d[0],&pLineA->picture);
					if (pLineA->pcbDrawPicture!=NULL)
					{
						pLineA->pcbDrawPicture(pLineA->contextDrawPicture,&pLineA->picture,pVM68k->d[1]);	
					}
				}
			}
			break;
		case 0xa0f1:
			{	// skip some words in the input buffer
				tVM68k_ubyte*	inputptr;
				tVM68k_uword	inputidx;
				tVM68k_ubyte	cinput;
				int i,n;
				inputptr=&pVM68k->pMem[pVM68k->a[1]&0xffff];
				inputidx=0;
				n=(pVM68k->d[0])&0xffff;
				for (i=0;i<n;i++)
				{
					do
					{
						cinput= READ_INT8BE(inputptr,inputidx++);
					} while (cinput);
				}
				pVM68k->a[1]+=inputidx;
			}
			break;
		case 0xa0f2:
			{
				tVM68k_uword objectnum;
				tProperties properties;
				int n;
				tVM68k_bool found;
				objectnum=(pVM68k->d[2])&0x7fff;
				n=pVM68k->d[4]&0x7fff;
				pVM68k->d[0]&=0xffff0000;
				pVM68k->d[0]|=pVM68k->d[2]&0xffff;
				found=0;
				retval=lineA_loadproperties(pLineA,pVM68k,objectnum,&pVM68k->a[0],&properties);
				do	
				{
					retval=lineA_loadproperties(pLineA,pVM68k,objectnum-1,NULL,&properties);
					if (properties.endflags&0x3fff)  found=1;
					if (objectnum==n) found=1;
					else objectnum--;
				} while ((objectnum!=0) && !found);
				if (found) pVM68k->sr|=(1<<0);            // bit 0 is the cflag
				pVM68k->d[2]&=0xffff0000;
				pVM68k->d[2]|=objectnum&0xffff;
			}
			break;
		case 0xa0f3: 
			if (pLineA->pcbOutputChar!=NULL)
			{
				pLineA->pcbOutputChar(pLineA->contextOutputChar,pVM68k->d[1],pVM68k->d[2]&0xff,pVM68k->d[3]&0xff);
			}
			break;
		case 0xa0f4:
			{
				char filename[64];
				int i;
				// TODO: version 0. The filename starts at A0.
				for (i=0;i<64;i++)	
				{
					filename[i]=pVM68k->pMem[(pVM68k->a[0]&0xffff)+i];
				}
				filename[63]=0;
			//	lineA_savegame(pLineA,pVM68k,filename);	// TODO: the way i am doing this is the brutal method. I am dumping the full memory.
				if (pLineA->pcbSaveGame!=NULL)
				{
					pLineA->pcbSaveGame(pLineA->contextSaveGame,
						&pVM68k->pMem[(pVM68k->a[0]&0xffff)],	// filename
						&pVM68k->pMem[(pVM68k->a[1]&0xffff)],	// ptr
						(pVM68k->d[1]&0xffff)	// len
					);
				}	
}
			break;
		case 0xa0f5:
			{
				char filename[64];
				int i;
				// TODO: version 0. The amount of bytes is stored in D1. The filename starts at A1.
				for (i=0;i<64;i++)	
				{
					filename[i]=pVM68k->pMem[(pVM68k->a[0]&0xffff)+i];	
				}
				filename[63]=0;
//				lineA_loadgame(pLineA,pVM68k,filename);	// TODO: the way i am doing this is the brutal method. I am dumping the full memory.
				if (pLineA->pcbLoadGame!=NULL)
				{
					pLineA->pcbLoadGame(pLineA->contextLoadGame,
						&pVM68k->pMem[(pVM68k->a[0]&0xffff)],	// filename
						&pVM68k->pMem[(pVM68k->a[1]&0xffff)],	// ptr
						(pVM68k->d[1]&0xffff)	// len
					);
				}	
			}
			break;
		case 0xa0f6:	// get random number (word), modulo D1.
			{
				tVM68k_ulong rand;
				tVM68k_uword limit;
				rand=lineA_getrandom(pLineA);	// advance the random generator
				limit=(pVM68k->d[1])&0xffff;
				if (limit==0) limit=1;
				rand%=limit;
				pVM68k->d[1]&=0xffff0000;
				pVM68k->d[1]|=(rand&0xffff);
			}
			break;
		case 0xa0f7:
			{	// get a random value between 0 and 255, and write it to D0.
				tVM68k_ulong rand;
				tVM68k_uword limit;
				rand=lineA_getrandom(pLineA);	// advance the random generator

				pVM68k->d[0]&=0xffffff00;
				pVM68k->d[0]|=((rand+(rand>>8))&0xff);
			}
			break;
		case 0xa0f8:	// write string
			{
				// strings are huffman-coded.
				// version 0: 'string2' holds the decoding tree in the first 256 bytes.
				// and the offset addresses for the bit streams in string1.
				// modes have bit 7 set.
				//
				// when the string is terminated with a \0 it ends.
				// when the string terminates with the sequence " @", it will be
				// extended. 
				//
				// the extension will have the cflag set.
				//
				tVM68k_uword idx;
				tVM68k_uword tmp;
				tVM68k_ubyte val;
				tVM68k_ubyte prevval;
				char string[1024];	// TODO
				int stringidx;
				char c;
				if (!(pVM68k->sr&(1<<0)))	// cflag is in bit 0.
				{
					pLineA->byteidx=0;	// start the byteidx and bitidx counter
					pLineA->bitidx=0;
					idx=pVM68k->d[0];
					if (idx==0) pLineA->byteidx=idx;
					// version 0: string 2 holds the table to decode the strings.
					// the decoder table is 256 bytes long. afterwards, a bunch of pointers
					// to bit indexes follow.
					else pLineA->byteidx=READ_INT16BE(pLineA->pStrings2,(0x100+2*idx));
					tmp=READ_INT16BE(pLineA->pStrings2,0x100);	
					if (tmp) 	// wrong version
					{
						fprintf(stderr,"FIXME. STRING1/STRING2\n");
						exit(0);
					}
				}
				val=0;
				stringidx=0;
				do
				{
					prevval=val;
					val=0;
					while (!(val&0x80))	// terminal symbols have bit 7 set.
					{
						tVM68k_ubyte bit;
						bit=pLineA->pStrings1[pLineA->byteidx];
						if (bit>>(pLineA->bitidx)&1)
						{
							val=pLineA->pStrings2[0x80+val];	// =1 -> go to the right
						} else {
							val=pLineA->pStrings2[     val];	// =0 -> go to the left
						}
						pLineA->bitidx++;
						if (pLineA->bitidx==8)
						{
							pLineA->bitidx=0;
							pLineA->byteidx++;
						}
					}
					val&=0x7f;	// remove bit 7.
					c=val;	// TODO: code conversion?
					string[stringidx++]=c;
			
				}
				while (val!=0 && (prevval!=' ' || val!='@'));	// end markers for the string are \0 and " @"
				if (prevval==' ' && val=='@')		// extend the string next time this function is being called.
				{
					pVM68k->sr|=(1<<0);	// set the cflag. cflag=bit 0.	
				} else {
					pVM68k->sr&=~(1<<0);	// clear the cflag. cflag=bit 0.	
				}
				string[stringidx]=0; // string has been assembled. and is 0 terminated now print it.
				// perform the callback so that the string becomes visible.	TODO: extended strings might be a problem!!
				if (pLineA->pcbOutputChar!=NULL)
				{
					pLineA->pcbOutputString(pLineA->contextOutputString,string,pVM68k->d[2]&0xff,pVM68k->d[3]&0xff);
				}
			}
			break;
		case 0xa0f9:	//get inventory item(d0)
			{
					// there is a list of parent objects
					//
					// apparently, the structure of the properties is as followed:
					// byte 0..4: UNKNOWN
					// byte 5: Flags. 
					//		bit 0: is_described
					// byte 6: some flags
					//		=bit 7: worn
					//		=bit 6: bodypart
					//		=bit 3: room
					//		=bit 2: hidden
					// byte 8/9: parent object. the player is =0x0000
					// byte 10..13: UNKNOWN
					// the data structure is a list.
				tVM68k_bool found;
				tVM68k_uword objectnum1;
				tVM68k_uword objectnum2;
				tProperties properties;

				found=0;
				// go backwards from the objectnumber
				for (objectnum1=pVM68k->d[0];objectnum1>0 && !found;objectnum1--)
				{
					objectnum2=objectnum1;
					do
					{
						// search for the parent
						retval=lineA_loadproperties(pLineA,pVM68k,objectnum2,&pVM68k->a[0],&properties);
						objectnum2=properties.parentobject;
						if ((properties.flags1&1) //is described
							|| (properties.flags2&0xcc))	// worn, bodypart, room or hidden
						{
							objectnum2=0;	// break the loop
						}
						else if (properties.parentobject==0) found=1;
						if (!properties.flags2&1) 
						{
							objectnum2=0;	// break the loop
						}
					} while (objectnum2);	
				}
				// set the z-flag when the object was found. otherwise clear it.
				pVM68k->sr&=~(1<<2);	// zflag is bit 2
				if (found) pVM68k->sr|=(1<<2);
				pVM68k->d[0]=objectnum1+1;	// return value
			}
			break;
		case 0xa0fa:
			{
				// search the properties database for a match with the entry in D2.
				// starting adress is stored in A0. D3 is the variable counter.
				// d4 is the limit. for (;D3<D4;D3++) {}
				// d5 =0 is a byte search. D5=1 is a word search.
				// set cflag when the entry is found.

				tVM68k_uword i;
				tVM68k_bool found;
				tVM68k_ulong addr;
				tVM68k_uword pattern;
				tVM68k_uword value;
				tVM68k_bool byte0word1;
	
				found=0;
				addr=pVM68k->a[0];
				pattern=pVM68k->d[2];
				byte0word1=pVM68k->d[5];
				pVM68k->sr&=~(1<<0);	// cflag is bit 0;	
				for (i=pVM68k->d[3];i<pVM68k->d[4] && !found;i++)
				{
					if (byte0word1)
					{
						value=READ_INT16BE(pVM68k->pMem,addr);
						value&=0x3fff;
					} else {
						value= READ_INT8BE(pVM68k->pMem,addr);
						value&=0xff;
					}
					addr+=14;
					if (value==pattern)
					{
						found=1;
						pVM68k->a[0]=addr;
						pVM68k->sr|=(1<<0);	// cflag is bit 0.
					}
				}
				pVM68k->d[3]=i;
				
			}
			break;
		case 0xa0fb:
           		{	// skip D2 many words in the dictionary, that is pointed at by A1
				tVM68k_ubyte*	dictptr;
				tVM68k_uword	dictidx;
				tVM68k_ubyte	cdict;
				int i;
				int n;
				

				dictidx=0;
				if (pLineA->version==0 || pLineA->pDict==NULL || pLineA->dictsize==0) 
				{
					dictptr=&pVM68k->pMem[pVM68k->a[1]&0xffff];
				} else {
					dictptr=pLineA->pDict;
				}
				n=pVM68k->d[2];
				for (i=0;i<n;i++)
				{
					do
					{
						cdict= READ_INT8BE(dictptr,dictidx++);
					} while (!(cdict&0x80));
				}
				pVM68k->d[2]&=0xffff0000;	// that was a counter
				pVM68k->a[1]+=dictidx;
			}
			break;
		case 0xa0fc:	// skip D0 many words in the input buffer, as well as the dictionary.
			{
				tVM68k_ubyte*	dictptr;
				tVM68k_ubyte*	inputptr;
				tVM68k_uword	dictidx;
				tVM68k_uword	inputidx;
				tVM68k_ubyte	cdict;
				tVM68k_ubyte	cinput;
				int i,n;
				dictidx=0;
				inputidx=0;
				if (pLineA->version==0 || pLineA->pDict==NULL || pLineA->dictsize==0) 
				{
					dictptr=&pVM68k->pMem[pVM68k->a[0]&0xffff];	// TODO: version 0. 
				} else {
					dictptr=pLineA->pDict;
				}
				inputptr=&pVM68k->pMem[pVM68k->a[1]&0xffff];
				n=pVM68k->d[0];		// FIXME: this cannot be right
				for (i=0;i<n&0xffff;i++)
				{
					while (!(dictptr[dictidx++]&0x80));	// in the dictionary, the end marker is bit 7 being set.
					while (inputptr[inputidx++]!=0x00);	// search for the end of the input.
				}
				pVM68k->d[0]&=0xffff0000;	// d0 was used as a counter
				pVM68k->a[0]+=dictidx;
				pVM68k->a[1]+=inputidx;
			}
			break;
		case 0xa0fd: 
				pLineA->properties_offset=pVM68k->a[0];
				if (pLineA->version!=0)
				{
					// version 1 introduced line F instructions
					pLineA->linef_subroutine=(pVM68k->a[3]&0xffff);
				}
				break;
		case 0xa0fe: 
				{
					// register D0 conatins an object number. calculate the address in memory
					tVM68k_sword objectnum;
					// TODO: this is how version 0 does it.
					pVM68k->d[0]&=0x00007fff;
					objectnum=pVM68k->d[0];
					pVM68k->a[0]=pLineA->properties_offset+objectnum*14;
				}
				break;
		case 0xa0ff:
				{
					// so, here's what i know: (version 0)
					// the dictonary is stored at A3. 
					// the word entered at A6
					// there is a "bank" in register D6
					//
					// the data structure is more or less plain. but the last char of each word has bit 7 set. 
					// special characters 0x81=ENDOFDICT 0x82=BANKSEPARATOR are used.
					//
					// input: (A6)
					// output: (A2)
					// dict: (A3)
					// objects: (A1)
					
					{
						tVM68k_ubyte* dtabptr;
						tVM68k_ubyte* inputptr;
						tVM68k_ubyte* outputptr;
						tVM68k_ubyte* dictptr;
						tVM68k_ubyte* objectptr;
						tVM68k_ubyte* adjptr;

						tVM68k_uword	inputidx;
						tVM68k_uword	outputidx;
						tVM68k_uword	outputidx2;
						tVM68k_uword	dictidx;
						tVM68k_uword	objectidx;
						tVM68k_uword	adjidx;
	
						tVM68k_uword	wordidx;
						tVM68k_ubyte	bank;
						tVM68k_ubyte	flag;
						tVM68k_bool	matching;
	
						tVM68k_ubyte	cinput1,cinput2;
						tVM68k_ubyte	cdict;
						tVM68k_bool	matchfound;
						tVM68k_ulong	wordmatch;
						tVM68k_ulong	tmp;
						int i,j;

						inputptr  =&pVM68k->pMem[pVM68k->a[6]];
						dtabptr   =&pVM68k->pMem[pVM68k->a[5]];	// version>0
						if (pLineA->version==0 || pLineA->pDict==NULL || pLineA->dictsize==0) 
						{
							dictptr=&pVM68k->pMem[pVM68k->a[3]&0xffff];
						} else {
							dictptr=pLineA->pDict;
						}
						outputptr =&pVM68k->pMem[pVM68k->a[2]];
						objectptr =&pVM68k->pMem[pVM68k->a[1]];
						adjptr    =&pVM68k->pMem[pVM68k->a[0]];
						inputidx=dictidx=objectidx=outputidx=adjidx=0;
						pVM68k->d[0]&=0xffff0000;		// this regsiter was used during the adjective search.
						pVM68k->d[1]&=0xffff0000;		// this regsiter was used during the adjective search.
				
						flag=0;
						bank=pVM68k->d[6];
						wordidx=0;
						cdict=0;
						matching=1;
						cinput1=cinput2=0;
						matchfound=0;
						// the way the first loop works is this:
						// character by character, a word from the dictionary is compared to the input.
						// when a mismatch happens, the beginning of the next word is searched. -> matching=0;
						// 
						while (cdict!=0x81)	// 0x81 is the end marker of the dictionary
						{
							cdict=dictptr[dictidx++];
							if (cdict==0x82)	// bank separator
							{
								inputidx=0;
								wordidx=0;
								bank++;
								matching=1;
							} else if (matching) {	// actively comparing
								cinput1=inputptr[inputidx++];	// the current character
								cinput2=inputptr[inputidx];	// and the next one

								if ((cinput1&0x5f)==(cdict&0x5f))	// the dictionary holds the same upper case character as the input
								{
									// end of an input word is either 0, space or ". end of a dictionary word has bit 7 set.
									if ((cinput2==0 || cinput2==' ' || cinput2=='\"')&&(cdict&0x80)) matchfound=1;
								} else {
									if (inputidx>7) matchfound=1;	// good enough. 
									matching=0;	// skip until the end of the word is reached.
								}
							}
							if (matchfound)
							{
								// the matches are stored in the following format:
								// bit 31..24 are a flag, which is =0 in version 0.
								// bit 23..16 is the bank.
								// bit 15..0 contain the matched word number in the bank.
								wordmatch =(((tVM68k_ulong)flag)<<24);
								wordmatch|=(((tVM68k_ulong)bank)<<16);
								wordmatch|=((tVM68k_ulong)wordidx);
#ifdef	DEBUG_PRINT
	{
		unsigned long long addr;

		addr=(unsigned long long)outputptr-(unsigned long long)pVM68k->pMem+outputidx;
//		printf("\n\nMEMWRITE LONG %08X @ %04llx\n",wordmatch,addr);
		printf("\nMEMWRITE WORD %04X @ %04llx\n",wordmatch&0xffff,addr+2);
		if ((wordmatch&0xffff)==0x0005) 
			printf("breakme\n");
	}
#endif
								WRITE_INT32BE(outputptr,outputidx,wordmatch);	// store the candidates in the output location.
								outputidx+=4;	// length of the result: 4 bytes.
								matchfound=0;
							}
							if (cdict&0x80 && cdict!=0x82)	// when the end of the word is reached. bit 7 is set.
							{
								wordidx++;
								matching=1;	// start over
								inputidx=0;	// start over.
							}
						}
#ifdef	DEBUG_PRINT
	{
		unsigned long long addr;

		addr=(unsigned long long)outputptr-(unsigned long long)pVM68k->pMem+outputidx;
		printf("\n\nMEMWRITE WORD %04X @ %04llx\n",0xffff,addr);
		
	}
#endif
	
						WRITE_INT16BE(outputptr,outputidx,0xffff);// the end marker in the buffer is a 0xffff.
						// the output buffer holds outputidx/4 many results.
						if (pLineA->version!=0)	// version 1 introduced synonyms.
						{
							// search the list of output words
							for (i=0;i<outputidx;i+=4)
							{
								wordmatch=READ_INT32BE(outputptr,i);
								flag=(wordmatch>>24)&0xff;
								bank=(wordmatch>>16)&0xff;
								wordidx=wordmatch&0xffff;
								if (bank==0x0b)
								{
									tVM68k_uword substword;
									substword=READ_INT16BE(dtabptr,wordidx*2);		// TODO: version >1???
		

									// the lower 5 bits are the bank.
									// the upper 11 bits in the substitute database are the actual word index.
									bank=substword&0x1f;
									wordidx=substword>>5;
									wordmatch=flag;wordmatch<<8;
									wordmatch|=(bank&0xff);wordmatch<<=16;
									wordmatch|=wordidx&0xffff;
									WRITE_INT32BE(outputptr,i,wordmatch);
							
								}

							} 
						}

						pVM68k->d[0]&=0xffff0000;
						outputidx2=0;
						adjidx=0;
						for (i=0;i<outputidx;i+=4)
						{
							tVM68k_uword	adjidx_base;
							tVM68k_uword	obj;
							tVM68k_bool	match;
							wordmatch=READ_INT32BE(outputptr,i);
							objectidx=0;
							flag=(wordmatch>>24)&0xff;
							bank=(wordmatch>>16)&0xff;
							wordidx=wordmatch&0xffff;
						
							match=1;	
							obj=READ_INT16BE(objectptr,objectidx);
							if (obj && bank==6)
							{
								// first step: skip the adjectives that are not meant for this word. each adjective list is separated by a 0.
								for (j=0;j<wordidx;j++)
								{
									do
									{
										cdict=READ_INT8BE(adjptr,adjidx++);
									} while (cdict!=0);
								}
								adjidx_base=adjidx;	// remeber the beginning of the list of adjectives for this word.

								match=0;
								while (!match && obj)
								{
									adjidx=adjidx_base;
									obj=READ_INT16BE(objectptr,objectidx);
									cinput1=(obj&0xff);	// given adjective
									if (obj)
									{
										objectidx+=2;
										do	// see if it matches any of the legal adjectives.
										{
											cdict=READ_INT8BE(adjptr,adjidx++);
											if ((cinput1+3)==cdict) match=1;
										} while (!match && cdict);
									}
								}
								adjidx=0;
							}
							
							pVM68k->d[1]&=0xffff0000;
							if (match==1) 
							{
								// TODO: version 0 does something with the flag
#ifdef	DEBUG_PRINT
								{
									unsigned long long addr;

									addr=(unsigned long long)outputptr-(unsigned long long)pVM68k->pMem+outputidx2+2;
									printf("\n\nMEMWRITE WORD %04X @ %04llx\n",wordmatch&0xffff,addr);

								}
#endif
		
								WRITE_INT32BE(outputptr,outputidx2,wordmatch);
								outputidx2+=4;
							} else pVM68k->d[1]|=1;
						}
						pVM68k->a[2]+=outputidx2;
//						pVM68k->d[0]=0;
						tmp=pVM68k->a[6];
						pVM68k->a[6]=pVM68k->a[6]+1;
						pVM68k->a[5]=tmp;
					}
				}
				break;
		default:
				printf("\n \x1b[0;37;44mUNIMPLEMENTED LINEA opcode %04X\x1b[0m\n",opcode);
				break;
	}

	return retval;	
}

int lineA_setCBoutputChar(void* hLineA,cbLineAOutputChar pCB,void *context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbOutputChar=pCB;
	pLineA->contextOutputChar=context;

	return LINEA_OK;
}
int lineA_setCBoutputString(void* hLineA,cbLineAOutputString pCB,void* context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbOutputString=pCB;
	pLineA->contextOutputString=context;

	return LINEA_OK;

}
int lineA_setCBinputString(void* hLineA,cbLineAInputString pCB,void* context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbInputString=pCB;
	pLineA->contextInputString=context;

	return LINEA_OK;

}
int lineA_setCBDrawPicture(void* hLineA,cbLineADrawPicture pCB,void* context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbDrawPicture=pCB;
	pLineA->contextDrawPicture=context;

	return LINEA_OK;

}

int lineA_setCBLoadGame(void* hLineA,cbLineALoadGame pCB,void* context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbLoadGame=pCB;
	pLineA->contextLoadGame=context;

	return LINEA_OK;

}

int lineA_setCBSaveGame(void* hLineA,cbLineASaveGame pCB,void* context)
{
	tLineA* pLineA=(tLineA*)hLineA;
	if (pLineA==NULL) return LINEA_NOK_INVALID_PTR;
	if (pLineA->magic!=MAGICVALUE) return LINEA_NOK_INVALID_PARAM;

	pLineA->pcbSaveGame=pCB;
	pLineA->contextSaveGame=context;

	return LINEA_OK;

}


